//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_PLATOONCREATEMESSAGE_H
#define PROTOCOL_PLATOONCREATEMESSAGE_H

#include <cstdint>
#include <vector>
#include "Protocol.h"
#include "json.hpp"
#include <NetworkingLib/Message.h>
#include <unordered_set>

namespace platoonProtocol
{

namespace messageTypes
{

constexpr MessageType LV_BROADCAST = 0x00000001;
constexpr MessageType FV_HEARTBEAT = 0x00000002;
constexpr MessageType LV_REQUEST = 0x00000004;
constexpr MessageType FV_REQUEST = 0x00000008;
constexpr MessageType ACCEPT_RESPONSE = 0x00000010;
constexpr MessageType REJECT_RESPONSE = 0x00000020;
constexpr MessageType LEAVE_PLATOON = 0x00000040;

const std::unordered_set<MessageType> standardMessageTypes{
    LV_BROADCAST, FV_HEARTBEAT, LV_REQUEST, FV_REQUEST, ACCEPT_RESPONSE, REJECT_RESPONSE, LEAVE_PLATOON};

inline bool isStandardMessageType(MessageType messageType)
{
    return messageTypes::standardMessageTypes.find(messageType) != messageTypes::standardMessageTypes.end();
}

}

namespace jsonNames
{

constexpr char VEHICLE_ID[] = "vehicle_id";
constexpr char PLATOON_ID[] = "platoon_id";
constexpr char PLATOON_SPEED[] = "ps";
constexpr char INNER_PLATOON_DISTANCE[] = "ips";

}

struct PlatoonMessage
{
    static constexpr std::size_t MAX_SIZE = 256;

    PlatoonMessage()
    {}

    PlatoonMessage(
        VehicleId vehicleId,
        MessageType messageType,
        PlatoonId platoonId,
        PlatoonSpeed platoonSpeed,
        InnerPlatoonDistance innerPlatoonDistance)
        : vehicleId(vehicleId)
          , messageType(messageType)
          , platoonId(platoonId)
          , platoonSpeed(platoonSpeed)
          , innerPlatoonDistance(innerPlatoonDistance)
    {}

    static PlatoonMessage followerRequest(VehicleId vehicleId)
    { return PlatoonMessage(vehicleId, messageTypes::FV_REQUEST, 0, 0, 0); }

    static PlatoonMessage leaderRequest(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::LV_REQUEST, platoonId, 0, 0); }

    static PlatoonMessage rejectResponse(VehicleId vehicleId)
    { return PlatoonMessage(vehicleId, messageTypes::REJECT_RESPONSE, 0, 0, 0); }

    static PlatoonMessage acceptResponse(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::ACCEPT_RESPONSE, platoonId, 0, 0); }

    static PlatoonMessage broadcastMessage(VehicleId vehicleId,
                                           PlatoonId platoonId,
                                           PlatoonSpeed platoonSpeed,
                                           InnerPlatoonDistance innerPlatoonDistance)
    { return PlatoonMessage(vehicleId, messageTypes::LV_BROADCAST, platoonId, platoonSpeed, innerPlatoonDistance); }

    static PlatoonMessage heartbeatMessage(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::FV_HEARTBEAT, platoonId, 0, 0); }

    static PlatoonMessage leaveMessage(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::LEAVE_PLATOON, platoonId, 0, 0); }

    VehicleId vehicleId;
    MessageType messageType;
    PlatoonId platoonId;
    PlatoonSpeed platoonSpeed;
    InnerPlatoonDistance innerPlatoonDistance;
};

}

namespace networking
{
namespace message
{

template<>
struct Encoder<platoonProtocol::PlatoonMessage>
{
    void operator()(const platoonProtocol::PlatoonMessage & message, std::string & data) const
    {
        using namespace platoonProtocol;

        auto messageType = message.messageType;

        if (!messageTypes::isStandardMessageType(messageType))
            throw 0;    // Throw something to signal a failed encoding.

        // 4 bytes for message type
        data = std::string(4, '\0');

        std::uint8_t numDataBytes[sizeof(MessageType)];
        networking::utils::toBigEndian<4>(numDataBytes, messageType);

        for (std::size_t i = 0; i < sizeof(MessageType); i++)
            data[i] = numDataBytes[i];

        using json = nlohmann::json;
        using namespace jsonNames;
        auto j = json{{VEHICLE_ID, message.vehicleId},
                      {PLATOON_ID, message.platoonId}};

        if (messageType == messageTypes::LV_BROADCAST)
        {
            j[PLATOON_SPEED] = message.platoonSpeed;
            j[INNER_PLATOON_DISTANCE] = message.innerPlatoonDistance;
        }

        data.append(j.dump());
    }
};

template<>
struct Decoder<platoonProtocol::PlatoonMessage>
{
    void operator()(platoonProtocol::PlatoonMessage & message, const std::string & data) const
    {
        using namespace platoonProtocol;

        if (data.size() < sizeof(MessageType))
            throw 0;    // Throw something to inform that we failed to decode the message.

        using json = nlohmann::json;
        auto j = json::parse(data.begin() + sizeof(MessageType), data.end());

        auto messageType = networking::utils::fromBigEndian<sizeof(MessageType), MessageType>(
            (const std::uint8_t *) data.c_str());

        // We only know message types conforming to the standard protocol specification.
        if (!messageTypes::isStandardMessageType(messageType))
            throw 0;

        message.messageType = messageType;

        using namespace jsonNames;
        message.vehicleId = j.at(VEHICLE_ID).get<VehicleId>();
        message.platoonId = j.at(PLATOON_ID).get<PlatoonId>();

        if (messageType == messageTypes::LV_BROADCAST)
        {
            message.platoonSpeed = j.at(PLATOON_SPEED).get<PlatoonSpeed>();
            message.innerPlatoonDistance = j.at(INNER_PLATOON_DISTANCE).get<InnerPlatoonDistance>();
        }
    };
};

}
}

#endif //PROTOCOL_PLATOONCREATEMESSAGE_H
